#!/bin/bash

. /etc/init.d/functions
. /usr/local/Reductor/etc/const
. $CONFIG

ssh_port=22
monitoring_port=8085

create_and_flush_chain() {
	chain="$1"
	table="${2:-filter}"
	iptables -t $table -N $chain 2>/dev/null || true
	iptables -t $table -F $chain
}

add_skip_check_rule() {
	if [ -s $HOOKDIR/skip_check.list ]; then
		echo  "- Добавляется правило DROP для белого списка абонентов"
		iptables -A FORWARD -m set --set skip_check src -j DROP
	fi
}

add_custom_rules_http() {
	if [ -s $HOOKDIR/custom_bindings ]; then
		local FORBIDDEN_TARGET="$1"
		echo "- Добавляются дополнительные правила фильтрации http"
		while read blacklist_id ipset_id redirect_url comment; do
			if [ "$ipset_id" = '-1' -a "$blacklist_id" = '0' ]; then
				echo "  - Удаляем дефолтное фильтрующее правило из-за записи в $HOOKDIR/custom_bindings: $blacklist_id $ipset_id $redirect_url $comment"
				chroot $CHROOTDIR /sbin/iptables -A 'http' -p tcp -m reductor --id $blacklist_id --comment "$comment" -j $FORBIDDEN_TARGET $redirect_url
				iptables -D http 1
				continue
			fi
			chroot $CHROOTDIR /sbin/iptables -A 'http' -p tcp -m set --match-set custom_$ipset_id src -m reductor --id $blacklist_id --comment "$comment" -j $FORBIDDEN_TARGET $redirect_url
		done < $HOOKDIR/custom_bindings
	fi
}

add_custom_rules_https() {
	if [ -s $HOOKDIR/custom_bindings_https ]; then
		local TARGET="REJECT --reject-with tcp-reset"
		echo "- Добавляются дополнительные правила фильтрации https"
		while read src dst _; do
			iptables -A https -p tcp -m set --match-set custom_$src src -m set --match-set custom_$dst dst -j $TARGET
		done < $HOOKDIR/custom_bindings_https
	fi
}

admin_notify() {
	local is_readonly
	local page="http://deny.carbonsoft.ru/whereismymoney"
	is_readonly="$(python -m json.tool /var/reg/CRB-Reductor.key | grep '"cf_regreadonly":' | egrep -o [0-9])" || true
	if [ "${misc['admin_notify']:-0}" = '1' -a "$is_readonly" = '1' ]; then
		echo "- Добавляем напоминание для администраторов о скорой блокировке"
		iptables -I http ! -d 95.213.143.204 -p tcp --tcp-flags PSH PSH -m set --set reductor_admins src -m limit --limit 5/hour -j FORBIDDEN --forbidden-url $page
	fi
}

http() {
	echo "- Настраиваем фильтрацию http"
	create_and_flush_chain "http"
	local TARGET="REJECT --reject-with tcp-reset"
	local FORBIDDEN_TARGET="FORBIDDEN --forbidden-url "
	if [ "${filter['forbidden']}" = '1' ]; then
		local url="${filter['forbidden_url']:-http://helpdesk.carbonsoft.ru/deny}"
		TARGET="FORBIDDEN"
		[ "${filter['save_domain']:-0}" == '1' ] && TARGET="$TARGET --save-domain"
		[ "${filter['forbidden_log']:-0}" == '1' ] && TARGET="$TARGET --log"
		FORBIDDEN_TARGET="$TARGET --forbidden-url" #for use in custom redirect
		TARGET="$FORBIDDEN_TARGET $url"
	fi
	local traffic_to_filter="-p tcp -m set --match-set http_ports dst"
	if [ -s $LISTDIR/our.white_http_iplist ]; then
		echo "- Добавляем белый список ip (http)"
		iptables -A http -m set --match-set white_http_reductor dst -j DROP
	fi
	chroot $CHROOTDIR /sbin/iptables -A 'http' -p tcp -m reductor --id 0 --comment 'rkn' -j $TARGET
	admin_notify
	add_custom_rules_http "$FORBIDDEN_TARGET"
	iptables -A FORWARD $traffic_to_filter -j http
}

https() {
	echo "- Настраиваем фильтрацию https"
	create_and_flush_chain "https"
	if [ -s $LISTDIR/our.white_https_iplist ]; then
		echo "- Добавляем белый список ip (https)"
		iptables -A https -p tcp --dport 443 -m set --match-set white_https_reductor dst -j DROP
	fi
	iptables -A https -p tcp -m set --match-set https_reductor dst -j REJECT --reject-with tcp-reset
	iptables -A https -p tcp -m set --match-set https_reductor src -j REJECT --reject-with tcp-reset
	add_custom_rules_https
	iptables -A FORWARD -p tcp --dport 443 -j https
	iptables -A FORWARD -p tcp --sport 443 -j https
}

ftp() {
	echo "- Настраиваем фильтрацию ftp"
	create_and_flush_chain "ftp"
	iptables -A ftp -p tcp -m set --match-set ftp_reductor dst -j REJECT --reject-with tcp-reset
	iptables -A ftp -p tcp -m set --match-set ftp_reductor src -j REJECT --reject-with tcp-reset
	iptables -A FORWARD -p tcp --dport 21 -j ftp
	iptables -A FORWARD -p tcp --sport 21 -j ftp
}

ip_port() {
	if [ "${filter['cardsharing']:-}" = '1' ]; then
		echo "- Настраиваем фильтрацию кардшаринга и прочего"
		create_and_flush_chain "other_traffic"
		for side in src dst; do
			iptables -A other_traffic -p tcp -m set --match-set ip_port_ban $side -m set --match-set tcp_ports $side -j REJECT --reject-with tcp-reset
			iptables -A other_traffic -p udp -m set --match-set ip_port_ban $side -m set --match-set udp_ports $side -j REJECT --reject-with icmp-port-unreachable
		done
		iptables -A FORWARD -j other_traffic
	fi
}

filter_policy() {
	iptables -P INPUT ACCEPT
	iptables -P FORWARD DROP
	iptables -P OUTPUT ACCEPT
}

dns() {
	if [ "${filter['dns']:-0}" = '1' ]; then
		local IP="${filter['dns_ip']:-127.0.0.1}"
		echo "- Настраиваем фильтрацию DNS"
		if [ "$IP" != "${IP%%:*}" ]; then
			$SETCOLOR_FAILURE
			echo "  - WARNING: обрезаем вероятно порт, указанный рядом с IP, DNS ничего не знает о портах"
			$SETCOLOR_NORMAL
			IP="${IP%%:*}"
		fi
		insert_module xt_dnsmatch
		insert_module ipt_DNSREPLY
		create_and_flush_chain "dns"
		iptables -I dns -p udp -m dnsmatch --id 0 --comment "rkn_by_domain" -j DNSREPLY --dnsreply-ip "$IP"
		iptables -A FORWARD -p udp --dport 53 -j dns
		if [ "${filter['dns_log_level']}" = '1' -o "${filter['dns_log_level']}" = '2' ]; then
			$SETCOLOR_FAILURE
			echo "  - WARNING: включаем DNS логирование, отключите его после завершения дебага"
			$SETCOLOR_NORMAL
			echo "${filter['dns_log_level']}" > /proc/net/ipt_dnsmatch/trace_ctl
		fi
	fi
}

filter() {
	filter_policy
	iptables -F
	iptables -A INPUT -m state --state ESTABLISHED,RELATED -j ACCEPT
	iptables -A INPUT -p icmp -j ACCEPT
	iptables -A INPUT -i lo -j ACCEPT
	if [ -n "${admin['ip']:-}" ]; then
		if [ "${filter['forbidden_url_local']:-0}" == '1' ]; then
			iptables -A INPUT -m state --state NEW -p tcp --dport 80 -j ACCEPT
			if [ "${filter['dns']:-0}" == '1' ]; then
				iptables -A INPUT -m state --state NEW -p tcp --dport 443 -j ACCEPT
			fi
		fi
		iptables -A INPUT -m state --state NEW -p tcp --dport "${admin['web_port']:-80}" -m set --match-set reductor_admins src -j ACCEPT
	fi
	iptables -A INPUT -m state --state NEW -m tcp -p tcp --dport $ssh_port -j ACCEPT
	iptables -A INPUT -j REJECT --reject-with icmp-host-prohibited

	add_skip_check_rule
	http
	https
	dns
	ftp
	ip_port
	ip_block
}

ip_block() {
	echo "- Настраиваем фильтрацию IP адресов"
	create_and_flush_chain "ip_block"
	iptables -A ip_block -p tcp -j REJECT --reject-with tcp-reset
	iptables -A ip_block -p udp -j REJECT --reject-with icmp-port-unreachable
	iptables -A ip_block -j REJECT --reject-with icmp-proto-unreachable
	iptables -A FORWARD -m set --match-set http_reductor dst,dst -j ip_block
}

mangle() {
	iptables -t mangle -F
	for chain in PREROUTING INPUT FORWARD OUTPUT POSTROUTING; do
		iptables -t mangle -P $chain ACCEPT
	done

	iptables -t mangle -A PREROUTING -m ttl --ttl-eq 1 -j DROP

	if [ "${misc['optimize_routing']}" = '1' ]; then
		:
		# iptables -t mangle -A PREROUTING -p tcp -m set --match-set http_ports dst -j ACCEPT
		# for ports in dports sports; do
		# 	iptables -t mangle -A PREROUTING -p tcp -m multiport --$ports "$([ "$ports" = 'sports' ] && echo "80,")443,$ssh_port,21,25,53,123,1488,$monitoring_port,8086,25251" -j ACCEPT
		# done
		# iptables -t mangle -A PREROUTING -p udp -m multiport --ports 53,123,161 -j ACCEPT
		# iptables -t mangle -A PREROUTING -i lo -j ACCEPT
		# iptables -t mangle -A PREROUTING -p icmp -j ACCEPT
		# iptables -t mangle -A PREROUTING -m set --match-set ip_port_reductor src,src -j ACCEPT
		# iptables -t mangle -A PREROUTING -m set --match-set ip_port_reductor dst,dst -j ACCEPT
		# iptables -t mangle -A PREROUTING -j DROP
	fi
}

nat() {
	iptables -t nat -F
	if [ "${filter['router']}" != '1' ]; then
		return 0
	fi
	echo "- Включается маршрутизация"
	iptables -A FORWARD -j ACCEPT
	{
		iptables -t nat -D POSTROUTING -j MASQUERADE || true
		iptables -t nat -D POSTROUTING -j MASQUERADE || true
	} &>/dev/null
	iptables -t nat -I POSTROUTING -j MASQUERADE
}

raw() {
	iptables -t raw -F PREROUTING
	for chain in mirror_traffic ip_port; do
		create_and_flush_chain "$chain" "raw"
	done
	if [ "${filter['notrack']:-0}" == '1' ]; then
		for link in $(brctl show | egrep -o eth[0-9\.]+$); do
			iptables -t raw -I PREROUTING -i $link -j mirror_traffic
		done
	fi
	iptables -t raw -A mirror_traffic -j NOTRACK
	if [ "${misc['optimize_routing']}" = '1' ]; then
		iptables -t raw -A mirror_traffic -p tcp -m set --match-set http_ports dst -j ACCEPT
		# iptables -t raw -A mirror_traffic -p tcp --dport 443 --tcp-flags SYN SYN -j ACCEPT
		iptables -t raw -A mirror_traffic -p tcp --dport 443 -j ACCEPT
		if [ "${filter['dns']:-0}" = '1' ]; then
			iptables -t raw -A mirror_traffic -p udp --dport 53 -j ACCEPT
		fi
		iptables -t raw -A mirror_traffic -m set --match-set ip_port_ban dst -j ACCEPT
		iptables -t raw -A mirror_traffic -m set --match-set http_reductor dst -j ACCEPT
		iptables -t raw -A mirror_traffic -j DROP
	fi

}

ipset_recreate() {
	local name="$1"
	ipset save | grep -q "^create $*" && return 0
	if ipset save | grep -q "^create $name"; then
		ipset flush "$name"
		ipset destroy "$name" || true
	fi
	echo "  - Пересоздаём $name"
	ipset create "$@"
}

reductor_admins() {
	# ok, no one reads a help text :(
	admin['ip']="${admin['ip']//[,;]/ }"
	ipset flush reductor_admins
	for ip in ${admin['ip']:-}; do
		ipset add reductor_admins $ip
	done
}

__ipset_config() {
	echo reductor_admins hash:net family inet hashsize 256 maxelem 512
	echo https_reductor hash:net family inet hashsize 8192 maxelem 65536
	if [ "${filter['ipv6']:-0}" = '1' ]; then
		echo https_reductor_ipv6 hash:net family inet6 hashsize 4096 maxelem 65536
	fi
	for ipset in http_reductor white_https_reductor white_http_reductor ftp_reductor skip_check; do
		echo "$ipset" hash:net family inet hashsize 1024 maxelem 65536
	done
	echo http_ports bitmap:port range 0-65535
	echo tcp_ports bitmap:port range 0-65535
	echo udp_ports bitmap:port range 0-65535
	echo ip_port_ban hash:ip family inet hashsize 1024 maxelem 2048
	echo ip_port_reductor hash:ip,port family inet hashsize 1024 maxelem 65536
	if [ -s $HOOKDIR/custom_bindings ]; then
		mkdir -p $HOOKDIR/custom_ipsets/
		for ipset in $(find $HOOKDIR/custom_ipsets/ -type f); do
			echo custom_${ipset##*/} hash:net family inet hashsize 1024 maxelem 65536
		done
	fi
}

ipset_state() {
	for ipset in $(ipset_config | awk '{print $1}'); do
		ipset save | grep "^create $ipset" | sed 's/create //; s/ $//'
	done | LANG='' sort -u
}

ipset_config() {
	__ipset_config | LANG='' sort -u
}

ipset_create() {
	ipset_config | while read params; do
		ipset_recreate $params
	done
}

skip_check() {
	if [ -s $HOOKDIR/skip_check.list ]; then
		echo "- Заполняем список пропуска проверок"
		ipset flush skip_check
		while read ip _; do
			if [ "${ip:0:1}" != '#' ]; then
				ipset add skip_check $ip
			fi
		done < $HOOKDIR/skip_check.list
	fi
}

custom_bindings() {
	if [ -s $HOOKDIR/custom_bindings ]; then
		mkdir -p $HOOKDIR/custom_ipsets/
		for ipset in $(find $HOOKDIR/custom_ipsets/ -type f); do
			ipset flush custom_${ipset##*/}
			while read ip _; do
				ipset add custom_${ipset##*/} $ip || true
			done < $ipset
		done
	fi
}

ipset_prepare() {
	ipset_create
	skip_check
	custom_bindings
	reductor_admins
}

ipv6() {
	echo "- Настраиваем фильтрацию IPv6"
	local traffic_to_filter="-p tcp -m set --match-set http_ports dst"
	local TARGET="REJECT --reject-with tcp-reset"
	if [ "${filter['forbidden']}" = '1' ]; then
		local url="${filter['forbidden_url']:-http://helpdesk.carbonsoft.ru/deny}"
		TARGET="FORBIDDEN --forbidden-url $url"
		if [ "${filter['save_domain']:-0}" == '1' ]; then
			TARGET="$TARGET --save-domain"
		fi
		if [ "${filter['forbidden_log']:-0}" == '1' ]; then
			TARGET="$TARGET --log"
		fi
	fi

	echo 0 > /proc/sys/net/ipv6/conf/all/disable_ipv6
	echo 0 > /proc/sys/net/ipv6/conf/default/disable_ipv6
	echo 1 > /proc/sys/net/ipv6/conf/all/forwarding
	modprobe ip6_tables
	modprobe ip6table_mangle
	modprobe ip6table_filter
	insert_module ip6t_FORBIDDEN
	ip6tables -F
	ip6tables -P FORWARD DROP

	chroot $CHROOTDIR /sbin/ip6tables -A FORWARD $traffic_to_filter -m reductor --id 0 --comment 'rkn' -j $TARGET
	ip6tables -A FORWARD -p tcp -m set --match-set https_reductor_ipv6 dst -j REJECT --reject-with tcp-reset
	ip6tables -A FORWARD -p tcp -m set --match-set https_reductor_ipv6 src -j REJECT --reject-with tcp-reset
	ebtables -t broute -A BROUTING -p IPv6 -j redirect  --redirect-target DROP
}

ipset_need_recreate() {
	# 0 - reload, 1 - skip
	ipset_config > $TMPDIR/ipset_config
	ipset_state > $TMPDIR/ipset_state
	! cmp $TMPDIR/ipset_config $TMPDIR/ipset_state
}

use_hook() {
	hook="$HOOKDIR/${0##*/}"
	if [ -x "$hook" ]; then
		. "$hook" __include
	fi
}

modules_checksum() {
	find $CHROOTDIR/{lib/modules/,lib64/xtables/} -type f -name "*.[sk]o" -exec md5sum  {} \; | md5sum
}

modules_need_reload() {
	# 0 - reload, 1 - skip
	local checksum_file="/var/lib/reductor/modules_checksum"
	local new_checksum
	new_checksum="$(modules_checksum)"
	mkdir -p /var/lib/reductor/
	touch $checksum_file
	[ ! -s $checksum_file ] && echo "$new_checksum" > $checksum_file && return 0
	read old_checksum < $checksum_file
	echo "$new_checksum" > $checksum_file
	[ "$new_checksum" != "$old_checksum" ]
}

start() {
	echo "- Конфигурация файрвола"
	insert_module xt_reductor
	[ "${filter['forbidden']}" = '1' ] && insert_module ipt_FORBIDDEN
	ipset_prepare
	raw
	mangle
	nat
	filter
	ebtables-restore < $MAINDIR/contrib/etc/sysconfig/ebtables
	[ "${filter['ipv6']:-0}" = '1' ] && ipv6
	[ -x $HOOKDIR/firewall.sh ] && $HOOKDIR/firewall.sh start || true
	echo "- Конфигурация файрвола завершена"
}

use_hook
case "${1//[_-]/}" in
	start )
		trap __exit EXIT
		set -eu
		"${1//[_-]/}"
		;;
	ipsetconfig )
		ipset_config
		;;
	ipsetstate )
		ipset_state
		;;
	ipsetneedrecreate )
		ipset_need_recreate
		;;
	modulesneedreload )
		modules_need_reload
		;;
	* )
		echo "$1 not supported, avaible options:"
		grep -A 1000 "case" $0 | grep '[a-z] )' | egrep -o [a-z]+ | sed 's/^/--/g'
		exit 1
		;;
esac
